from hashformers.segmenter.data_structures import (
    WordSegmenterOutput,
    TweetSegmenterOutput
)

from collections.abc import Iterable

def coerce_segmenter_objects(method):
    """
    A decorator function that ensures the returned value from the decorated method is of certain types 
    or converts the returned value to one of the allowed types.

    It also handles different types of 'inputs' passed to the decorated method. It checks whether the input is a string 
    or an iterable, and in the case of unsupported input type, raises a NotImplementedError.

    Args:
        method (function): The method to be decorated.

    Returns:
        function: The decorated function which enforces specific output types and handles different input types.
    """
    def wrapper(self, inputs, *args, **kwargs):
        
        if isinstance(inputs, str):
            output = method(self, [inputs], *args, **kwargs)
        elif isinstance(inputs, Iterable):
            output = method(self, inputs, *args, **kwargs)
        else:
            raise NotImplementedError(str(type(inputs)))
        
        for allowed_type in [
            WordSegmenterOutput,
            TweetSegmenterOutput
        ]:
            if isinstance(output, allowed_type):
                return output
        
        if isinstance(output, str):
            return WordSegmenterOutput(output=[output])

        if isinstance(output, Iterable):
            return WordSegmenterOutput(output=output)

    return wrapper

class BaseSegmenter(object):
    """
    Base class for text segmenter objects.
    """
    @coerce_segmenter_objects
    def predict(self, inputs, *args, **kwargs):
        """
        Predict method that delegates to the segment method.
        It is decorated with coerce_segmenter_objects to handle different input types and enforce output type.

        Args:
            inputs (str or Iterable): The inputs to be segmented.
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.

        Returns:
            WordSegmenterOutput or TweetSegmenterOutput: The output from the segment method.
        """
        return self.segment(inputs, *args, **kwargs)
    
    def segment(self, inputs, *args, **kwargs):
        """
        Abstract method for segmentation. Should be implemented in a child class.

        Args:
            inputs (str or Iterable): The inputs to be segmented.
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.

        Raises:
            NotImplementedError: If this method is not overridden in a child class.
        """
        raise NotImplementedError("This method should be implemented in a child class.")

    def preprocess(self, inputs, lower=False, remove_hashtag=True, hashtag_character="#"):
        """
        Preprocesses the inputs based on the given parameters.

        Args:
            inputs (str or Iterable): The inputs to be preprocessed.
            lower (bool, optional): Whether to convert the inputs to lower case. Defaults to False.
            remove_hashtag (bool, optional): Whether to remove the hashtag character from the inputs. Defaults to True.
            hashtag_character (str, optional): The hashtag character to be removed. Defaults to "#".

        Returns:
            str or list of str: The preprocessed inputs.

        Raises:
            NotImplementedError: If the type of inputs is neither str nor Iterable.
        """
        def preprocess_input(word):
            if lower:
                word = word.lower()
            if remove_hashtag:
                word = word.lstrip(hashtag_character)
            return word
        
        if isinstance(inputs, str):
            inputs = preprocess_input(inputs)
        elif isinstance(inputs, Iterable):
            inputs = [ preprocess_input(x) for x in inputs ]
        else:
            raise NotImplementedError(str(type(inputs)))
        
        return inputs


from typing import Union, Any
import pandas as pd
from dataclasses import dataclass

@dataclass
class WordSegmenterOutput:
    output: list[str]
    segmenter_rank: Union[pd.DataFrame, None] = None
    reranker_rank: Union[pd.DataFrame, None] = None
    ensemble_rank: Union[pd.DataFrame, None] = None

@dataclass
class HashtagContainer:
    hashtags: list[list[str]]
    hashtag_set: list[str]
    replacement_dict: dict

@dataclass
class TweetSegmenterOutput:
    output: list[str]
    word_segmenter_output: Any


from hashformers.beamsearch.data_structures import enforce_prob_dict
from typing import Any
from ttp import ttp
from hashformers.segmenter.base_segmenter import BaseSegmenter
from hashformers.segmenter.regex_segmenter import RegexWordSegmenter
from hashformers.segmenter.data_structures import ( 
    WordSegmenterOutput, 
    TweetSegmenterOutput, 
    HashtagContainer
)
import re
from functools import reduce
import dataclasses
import pandas as pd


class BaseWordSegmenter(BaseSegmenter):
    """
    Initializes BaseWordSegmenter class with segmenter, reranker and ensembler models.

    Args:
        segmenter: The model used for initial word segmentation.
        reranker: The model used for reranking the segmented words.
        ensembler: The model used for ensemble operations over the segmenter and reranker models.
    """
    def __init__(
        self,
        segmenter = None,
        reranker = None,
        ensembler = None
    ):
        self.segmenter_model = segmenter
        self.reranker_model = reranker
        self.ensembler = ensembler

    def get_segmenter(self):
        """
        Returns the segmenter model.
        """
        return self.segmenter_model.model

    def get_reranker(self):
        """
        Returns the reranker model.
        """
        return self.reranker_model.model

    def get_ensembler(self):
        """
        Returns the ensembler model.
        """
        return self.ensembler

    def set_segmenter(self, segmenter):
        """
        Sets the segmenter model.

        Args:
            segmenter: The model used for initial hashtag segmentation.
        """
        self.segmenter_model.model = segmenter
    
    def set_reranker(self, reranker):
        """
        Sets the reranker model.

        Args:
            reranker: The model used for reranking the segmented hashtags.
        """
        self.reranker_model.model = reranker

    def set_ensembler(self, ensembler):
        """
        Sets the ensembler model.

        Args:
            ensembler: The model used for ensemble operations over the segmenter and reranker models.
        """
        self.ensembler = ensembler

    def segment(
            self,
            word_list: list[str],
            segmenter_run: Any = None,
            preprocessing_kwargs: dict = {},
            segmenter_kwargs: dict = {},
            ensembler_kwargs: dict = {},
            reranker_kwargs: dict = {},
            use_reranker: bool = True,
            use_ensembler: bool = True,
            return_ranks: bool = False) -> Any :
        """
        Segments the input list of words using the segmenter, reranker, and ensembler models.
        Allows customization of the segmenting process with multiple keyword arguments.

        Args:
            word_list: List of strings, where each string is a word to be segmented.
            segmenter_run: Optional argument to use a pre-existing segmenter run, defaults to None.
            preprocessing_kwargs: Keyword arguments to be used during the preprocessing phase.
            segmenter_kwargs: Keyword arguments to be used by the segmenter model.
            ensembler_kwargs: Keyword arguments to be used by the ensembler model.
            reranker_kwargs: Keyword arguments to be used by the reranker model.
            use_reranker: Boolean flag to indicate whether to use the reranker model, defaults to True.
            use_ensembler: Boolean flag to indicate whether to use the ensembler model, defaults to True.
            return_ranks: Boolean flag to indicate whether to return the ranks from the models, defaults to False.

        Returns:
            Returns the segmented words. If return_ranks is True, also returns the segmenter_rank, reranker_rank, and ensemble_rank.
        """
        word_list = super().preprocess(word_list, **preprocessing_kwargs)

        if not isinstance(segmenter_run, pd.DataFrame):
            segmenter_run = self.segmenter_model.run(
                word_list,
                **segmenter_kwargs
            )
        
        ensemble_prob_dict = None

        segmenter_prob_dict = enforce_prob_dict(
                segmenter_run,
                score_field="score"
        )

        if use_reranker and self.reranker_model:
            reranker_run = self.reranker_model.rerank(segmenter_run, **reranker_kwargs)

        if use_reranker and self.reranker_model and use_ensembler and self.ensembler:
            ensemble_prob_dict = self.ensembler.run(
                segmenter_run,
                reranker_run,
                **ensembler_kwargs
            )
            segs = ensemble_prob_dict.get_segmentations(
                astype="list",
                gold_array=word_list
            )

        else:
            segs = segmenter_prob_dict.get_segmentations(
                astype="list",
                gold_array=word_list
            )

        if not return_ranks:
            return segs
        else:
            segmenter_df = segmenter_prob_dict.to_dataframe().reset_index(drop=True)
            reranker_df = None
            ensembler_df = None

            if use_reranker:
                if self.reranker_model:
                    reranker_df = reranker_run.to_dataframe().reset_index(drop=True)
                if use_ensembler and self.ensembler and ensemble_prob_dict: 
                    ensembler_df = ensemble_prob_dict.to_dataframe().reset_index(drop=True)

            return WordSegmenterOutput(
                segmenter_rank=segmenter_df,
                reranker_rank=reranker_df,
                ensemble_rank=ensembler_df,
                output=segs
            )

class TwitterTextMatcher(object):
    
    def __init__(self):
        """
        Initializes TwitterTextMatcher object with a Parser from ttp module.
        """
        self.parser = ttp.Parser()
    
    def __call__(self, tweets):
        """
        Makes the TwitterTextMatcher instance callable. It parses the given tweets and returns their tags.

        Args:
            tweets: A list of strings, where each string is a tweet.

        Returns:
            A list of hashtags for each tweet.
        """
        return [ self.parser.parse(x).tags for x in tweets ]
    
class TweetSegmenter(BaseSegmenter):

    def __init__(self, matcher=None, word_segmenter=None):
        """
        Initializes a TweetSegmenter instance with a TwitterTextMatcher and a WordSegmenter.

        Args:
            matcher (TwitterTextMatcher, optional): Instance of TwitterTextMatcher used for matching text in tweets. 
                Defaults to an instance of TwitterTextMatcher if not provided.
            word_segmenter (WordSegmenter, optional): Instance of WordSegmenter used for segmenting words in tweets. 
                Defaults to an instance of RegexWordSegmenter if not provided.
        """
        if matcher:
            self.matcher = matcher
        else:
            self.matcher = TwitterTextMatcher()

        if word_segmenter:
            self.word_segmenter = word_segmenter
        else:
            self.word_segmenter = RegexWordSegmenter()

    def extract_hashtags(self, tweets):
        """
        Extracts hashtags from the provided list of tweets.

        Args:
            tweets: A list of strings, where each string is a tweet.

        Returns:
            A list of hashtags extracted from each tweet.
        """
        return self.matcher(tweets)

    def compile_dict(self, hashtags, segmentations, hashtag_token=None, lower=False, separator=" ", hashtag_character="#"):
        """
        Compiles a dictionary mapping hashtags to their corresponding segmentations.

        Args:
            hashtags (list): List of hashtags extracted from tweets.
            segmentations (list): Corresponding segmentations of the hashtags.
            hashtag_token (str, optional): Token to prepend to the segmented hashtag value. If not provided, no token is prepended.
            lower (bool, optional): If True, converts the hashtag value to lowercase. Defaults to False.
            separator (str, optional): Separator used between hashtag_token and the value. Defaults to " ".
            hashtag_character (str, optional): Character representing a hashtag. Defaults to "#".

        Returns:
            dict: A dictionary mapping hashtags to their segmented versions.
        """
        hashtag_buffer = {
            k:v for k,v in zip(hashtags, segmentations)
        }

        replacement_dict = {}

        for key, value in hashtag_buffer.items():

            if not key.startswith(hashtag_character):
                hashtag_key = hashtag_character + key
            else:
                hashtag_key = key

            if hashtag_token:
                hashtag_value = hashtag_token + separator + value
            else:
                hashtag_value = value

            if lower:
                hashtag_value = hashtag_value.lower()

            replacement_dict.update({hashtag_key : hashtag_value})

        return replacement_dict

    def replace_hashtags(self, tweet, regex_pattern, replacement_dict):
        """
        Replaces the hashtags in a tweet based on a provided replacement dictionary.

        Args:
            tweet (str): The tweet in which hashtags are to be replaced.
            regex_pattern (re.Pattern): Compiled regular expression pattern for matching hashtags in the tweet.
            replacement_dict (dict): Dictionary with original hashtags as keys and their replacements as values.

        Returns:
            str: The tweet with hashtags replaced.
        """
        if not replacement_dict:
            return tweet

        tweet = regex_pattern.sub(lambda m: replacement_dict[m.group(0)], tweet)
        
        return tweet

    def segmented_tweet_generator(self, tweets, hashtags, hashtag_set, replacement_dict, flag=0):
        """
        Yields segmented tweets from a provided list of tweets.

        Args:
            tweets (list): List of tweets to be segmented.
            hashtags (list): List of hashtags extracted from each tweet.
            hashtag_set (set): Set of unique hashtags extracted from all tweets.
            replacement_dict (dict): Dictionary with original hashtags as keys and their replacements as values.
            flag (int, optional): Flags for the regular expression compilation. Defaults to 0.

        Yields:
            str: Segmented version of each tweet.
        """
        hashtag_set_index = { value:idx for idx, value in enumerate(hashtag_set)}
        replacement_pairs = [ (key, value) for key, value in replacement_dict.items() ]

        for idx, tweet_hashtags in enumerate(hashtags):
            
            tweet_dict = [ hashtag_set_index[hashtag] for hashtag in tweet_hashtags]
            tweet_dict = [ replacement_pairs[index] for index in tweet_dict ]
            tweet_dict = dict(tweet_dict)

            # Treats edge case: overlapping hashtags
            tweet_map = \
                map(re.escape, sorted(tweet_dict, key=len, reverse=True))
            
            regex_pattern = re.compile("|".join(tweet_map), flag)
            tweet = self.replace_hashtags(tweets[idx], regex_pattern, tweet_dict)
            yield tweet

    def build_hashtag_container(self, tweets: str, preprocessing_kwargs: dict = {}, segmenter_kwargs: dict = {} ):
        """
        Constructs a HashtagContainer from a list of tweets.

        Args:
            tweets (list): List of tweets.
            preprocessing_kwargs (dict, optional): Keyword arguments for preprocessing. Defaults to an empty dictionary.
            segmenter_kwargs (dict, optional): Keyword arguments for the segmenter. Defaults to an empty dictionary.

        Returns:
            tuple: A tuple containing a HashtagContainer instance and the output from the word segmenter.
        """        
        hashtags = self.extract_hashtags(tweets)

        hashtag_set = list(set(reduce(lambda x, y: x + y, hashtags)))
        
        word_segmenter_output = self.word_segmenter.predict(hashtag_set, **segmenter_kwargs)

        segmentations = word_segmenter_output.output

        replacement_dict = self.compile_dict(hashtag_set, segmentations, **preprocessing_kwargs)

        return HashtagContainer(hashtags, hashtag_set, replacement_dict), word_segmenter_output
  
    def segment(self, tweets: list[str], regex_flag: Any = 0, preprocessing_kwargs: dict = {}, segmenter_kwargs: dict = {} ):
        """
        Segments a list of tweets into individual words and replaces the hashtags based on the preprocessing and segmenter configurations.

        Args:
            tweets (List[str]): List of tweets to be segmented.
            regex_flag (Any, optional): Regular expression flags used in replacing the hashtags. Defaults to 0.
            preprocessing_kwargs (dict, optional): Dictionary of keyword arguments used for preprocessing the tweets. Defaults to an empty dictionary.
            segmenter_kwargs (dict, optional): Dictionary of keyword arguments used for the WordSegmenter. Defaults to an empty dictionary.

        Returns:
            TweetSegmenterOutput: Contains the output of WordSegmenter and the segmented tweets.
        """
        hashtag_container, word_segmenter_output = self.build_hashtag_container(tweets, preprocessing_kwargs, segmenter_kwargs)
        output = list(self.segmented_tweet_generator(tweets, *dataclasses.astuple(hashtag_container), flag=regex_flag))

        return TweetSegmenterOutput(
            word_segmenter_output = word_segmenter_output,
            output = output
        )


import pandas as pd
import copy
import numpy as np 

from hashformers.experiments.evaluation import (
    filter_top_k
)

def project_scores(a, b, segmentation_field="segmentation", score_field="score"):
    """
    Projects the score from dataframe 'b' onto dataframe 'a' based on the segmentation_field.

    It first creates a view of dataframe 'b' with unique values of the segmentation_field.
    Then it merges this view with dataframe 'a' and updates the score field in 'a' with the score from 'b'.
    The resulting dataframe is sorted by score in ascending order.

    Args:
        a (pandas.DataFrame): The dataframe onto which the scores are to be projected.
        b (pandas.DataFrame): The dataframe from which the scores are taken.
        segmentation_field (str, optional): The field based on which scores are projected. Defaults to "segmentation".
        score_field (str, optional): The field which contains the scores to be projected. Defaults to "score".

    Returns:
        pandas.DataFrame: The dataframe 'a' with updated scores projected from dataframe 'b'.
    """
    b_view = b[[segmentation_field, score_field]]\
        .drop_duplicates(subset=[segmentation_field])
    df = pd.merge(a, b_view, on=segmentation_field, how='left')
    df = df.drop([score_field+'_x'], axis=1)
    df = df.rename(columns={
        score_field+'_y': score_field
    })
    df = df.sort_values(by=score_field, ascending=True)
    return df

def filter_and_project_scores(a, b, characters_field="hashtag", segmentation_field="segmentation"):
    """
    Filters the top two records of the dataframe 'a', projects the scores from dataframe 'b' onto 'a',
    and returns both the modified dataframes.

    Args:
        a (pandas.DataFrame): The first dataframe, which will be filtered and onto which the scores will be projected.
        b (pandas.DataFrame): The second dataframe, from which the scores are taken.
        characters_field (str, optional): The field used to sort the dataframes. Defaults to "hashtag".
        segmentation_field (str, optional): The field based on which scores are projected. Defaults to "segmentation".

    Returns:
        list of pandas.DataFrame: The modified dataframes 'a' and 'b' after filtering and projecting scores.
    """
    models = copy.deepcopy([a,b])
    for idx, m in enumerate(models):
        models[idx] = models[idx]\
            .sort_values(by=[characters_field, segmentation_field])

    models[0] = filter_top_k(models[0], 2, fill=True)
    models[1] = project_scores(models[0], models[1])

    for idx, m in enumerate(models):
        models[idx] = models[idx]\
            .sort_values(by=[characters_field, segmentation_field])\
            .reset_index(drop=True)
    return models

def calculate_diff_scores(a, b, characters_field="hashtag", score_field="score"):
    """
    Calculates the difference in scores between pairs of records in the dataframes 'a' and 'b'.

    Args:
        a (pandas.DataFrame): The first dataframe.
        b (pandas.DataFrame): The second dataframe.
        characters_field (str, optional): The field used to sort the dataframes. Defaults to "hashtag".
        score_field (str, optional): The field which contains the scores. Defaults to "score".

    Returns:
        list of pandas.DataFrame: The modified dataframes 'a' and 'b' with an additional 'diff' column indicating the score difference.
    """
    models = copy.deepcopy([a,b])
    for idx, m in enumerate(models):
        
        models[idx] = models[idx]\
            .sort_values(by=[characters_field, score_field])
        score_pairs = models[idx][score_field].values.reshape(-1,2)

        models[idx]['rank'] = \
            score_pairs.argsort().flatten()
        models[idx]['diff'] = \
            np.repeat(np.subtract.reduce(score_pairs, axis=1).flatten(), 2)
        models[idx]['diff'] = \
            models[idx]['diff'].fillna(0.0)
    return models

def build_ensemble_df(a, b):
    """
    Builds an ensemble dataframe from the input dataframes 'a' and 'b'.

    It filters and projects the scores from 'b' onto 'a', calculates the score differences,
    and then merges the differences back into the 'a' dataframe. 

    Args:
        a (pandas.DataFrame): The first dataframe.
        b (pandas.DataFrame): The second dataframe.

    Returns:
        pandas.DataFrame: The resulting ensemble dataframe with projected scores and score differences.
    """
    models = filter_and_project_scores(a, b)
    models = calculate_diff_scores(models[0], models[1])
    
    for idx, m in enumerate(models):
        models[idx]['diff'] = np.abs(models[idx]['diff'].values)

    models[0]['diff_2'] = models[1]['diff'] 
    models[0]['rank_2'] = models[1]['rank']

    return models[0]


"""
spaCy integration for hashformers.

This module provides a spaCy pipeline component for word segmentation
using hashformers' TransformerWordSegmenter.

Example usage:
    >>> import spacy
    >>> from hashformers.spacy import HashformersComponent
    >>> 
    >>> nlp = spacy.blank("en")
    >>> nlp.add_pipe("hashformers", config={"model": "distilgpt2"})
    >>> 
    >>> doc = nlp("#weneedanationalpark")
    >>> print(doc._.segmented)  # "we need a national park"
"""

try:
    import spacy
    from spacy.language import Language
    from spacy.tokens import Doc
    SPACY_AVAILABLE = True
except ImportError:
    SPACY_AVAILABLE = False

from hashformers import TransformerWordSegmenter


class HashformersComponent:
    """
    A spaCy pipeline component that performs word segmentation on text.
    
    This component uses hashformers' TransformerWordSegmenter to segment
    text (typically hashtags) into individual words. The segmented text
    is stored in the doc._.segmented extension.
    
    Attributes:
        nlp: The spaCy Language object.
        name: The component name.
        segmenter: The TransformerWordSegmenter instance.
    """
    
    def __init__(
        self,
        nlp,
        name: str,
        model: str = "distilgpt2",
        device: str = "cuda",
        gpu_batch_size: int = 1000
    ):
        """
        Initialize the HashformersComponent.
        
        Args:
            nlp: The spaCy Language object.
            name: The component name in the pipeline.
            model: The Hugging Face model name or path for segmentation.
                Defaults to "distilgpt2".
            device: Device to run the model on ("cuda" or "cpu").
                Defaults to "cuda".
            gpu_batch_size: Batch size for GPU processing.
                Defaults to 1000.
        """
        self.nlp = nlp
        self.name = name
        self.segmenter = TransformerWordSegmenter(
            segmenter_model_name_or_path=model,
            segmenter_device=device,
            segmenter_gpu_batch_size=gpu_batch_size
        )
        
        # Register the custom extension if not already registered
        if not Doc.has_extension("segmented"):
            Doc.set_extension("segmented", default=None)
    
    def __call__(self, doc: "Doc") -> "Doc":
        """
        Process a spaCy Doc and add segmented text to doc._.segmented.
        
        The component strips any leading '#' character from the text
        before segmentation.
        
        Args:
            doc: The spaCy Doc object to process.
            
        Returns:
            The processed Doc with segmented text in doc._.segmented.
        """
        # Get the text, stripping any leading '#' characters
        text = doc.text.lstrip('#')
        
        if text:
            # Perform segmentation
            result = self.segmenter.segment([text])
            doc._.segmented = result[0] if result else text
        else:
            doc._.segmented = ""
        
        return doc


# Register the component with spaCy if spaCy is available
if SPACY_AVAILABLE:
    @Language.factory(
        "hashformers",
        default_config={
            "model": "distilgpt2",
            "device": "cuda",
            "gpu_batch_size": 1000
        }
    )
    def create_hashformers_component(
        nlp: Language,
        name: str,
        model: str,
        device: str,
        gpu_batch_size: int
    ) -> HashformersComponent:
        """
        Factory function to create a HashformersComponent for spaCy pipelines.
        
        Args:
            nlp: The spaCy Language object.
            name: The component name.
            model: The Hugging Face model name or path.
            device: Device to run the model on.
            gpu_batch_size: Batch size for GPU processing.
            
        Returns:
            A configured HashformersComponent instance.
        """
        return HashformersComponent(
            nlp=nlp,
            name=name,
            model=model,
            device=device,
            gpu_batch_size=gpu_batch_size
        )


__all__ = ["HashformersComponent"]



from hashformers.beamsearch.data_structures import (
    enforce_prob_dict,
    ProbabilityDictionary
)
from hashformers.beamsearch.model_lm import ModelLM

class Reranker(ModelLM):
    """
    A class that inherits from the ModelLM class and specializes in re-ranking given data.

    This class provides a method to re-rank data using a particular language model, which is determined
    during initialization from the parent ModelLM class. The re-ranking is performed based on the 
    probabilities produced by the language model.

    Args:
        model_name_or_path (str, optional): The name or path of the pre-trained model. Default is "bert-base-cased".
        model_type (str, optional): The type of the model to use. Default is "bert".
        gpu_batch_size (int, optional): The batch size to use when performing computations on the GPU. Default is 1000.
        gpu_id (int, optional): The ID of the GPU to use. Default is 0.
        device (str, optional): The device on which to run the computations. Default is "cuda".

    """
    def __init__(
        self,
        model_name_or_path="bert-base-cased",
        model_type="bert",
        gpu_batch_size=1000,
        gpu_id=0,
        device="cuda"
    ):
        super().__init__(
            model_name_or_path=model_name_or_path,
            model_type=model_type,
            device=device,
            gpu_batch_size=gpu_batch_size,
            gpu_id=gpu_id
        )
    
    def rerank(
        self,
        data
    ):
        """
        Reranks the given data using the language model.

        The data is first converted to a ProbabilityDictionary, and then the model's probabilities 
        for each candidate in the data are computed. The results are returned as a new 
        ProbabilityDictionary.

        Args:
            data (dict or ProbabilityDictionary): The data to be reranked. If a dict is provided, it is 
                converted to a ProbabilityDictionary.

        Returns:
            ProbabilityDictionary: A ProbabilityDictionary of the reranked data, where the keys are 
                the candidates from the input data and the values are the scores computed by the 
                language model.
        """
        input_data = enforce_prob_dict(data)
        candidates = list(input_data.dictionary.keys())
        scores = self.model.get_probs(candidates)
        rank = { k:v for k,v in list(zip(candidates, scores))}
        return ProbabilityDictionary(rank)


import itertools
import re
from hashformers.beamsearch.data_structures import (
    Node,
    ProbabilityDictionary
)

from hashformers.beamsearch.model_lm import ModelLM

class Beamsearch(ModelLM):

    def __init__(
            self,
            model_name_or_path="gpt2", 
            model_type="gpt2", 
            device='cuda', 
            gpu_batch_size=1000):
        """
        Initializes the Beamsearch class.

        Args:
            model_name_or_path (str): Name of the model or path to the model to be loaded. Default is "gpt2".
            model_type (str): Type of the model. Default is "gpt2".
            device (str): Device to be used for computation. Default is 'cuda'.
            gpu_batch_size (int): Size of the batch to be processed on the GPU. Default is 1000.
        """
        super().__init__(
            model_name_or_path=model_name_or_path, 
            model_type=model_type, 
            device=device, 
            gpu_batch_size=gpu_batch_size)

    def next_step(self, list_of_candidates):
        """
        Generates the next possible candidates.

        Args:
            list_of_candidates (List[str]): List of current candidate strings.
        
        Returns:
            List[str]: List of possible next candidates.
        """
        output = []
        for candidate_string in list_of_candidates:
            candidates = [ 
                candidate_string[:pos] + ' ' + candidate_string[pos:] \
                    if pos else candidate_string for pos in range(len(candidate_string)) 
                ]
            candidates = list(filter(lambda x: not re.findall(".*?(?=\s{2})", x), candidates))
            output.extend(candidates)
        return output

    def update_probabilities(self, tree, prob_dict):
        """
        Updates the probabilities in the given probability dictionary.

        Args:
            tree (List[str]): List of candidate strings.
            prob_dict (dict): Dictionary of probabilities of the candidates.
        
        Returns:
            dict: Updated probability dictionary.
        """
        for item in tree:
            current_batch = []
            for word in item:
                if word in prob_dict:
                    continue
                else:
                    current_batch.append(word)
            if current_batch:
                current_batch_probs = self.model.get_probs(current_batch)
            for idx, word in enumerate(current_batch):
                prob_dict[word] = current_batch_probs[idx]
        return prob_dict

    def reshape_tree(self, tree, measure):
        """
        Reshapes the tree according to the provided measure.

        Args:
            tree (List[str]): List of candidate strings.
            measure (int): Measure to reshape the tree.
        
        Returns:
            List[List[str]]: Reshaped tree.
        """
        return [ tree[x:x+measure] for x in range(0, len(tree), measure) ]

    def flatten_list(self, list_):
        """
        Flattens a nested list.

        Args:
            list_ (List[List[Any]]): Nested list to be flattened.
        
        Returns:
            List[Any]: Flattened list.
        """
        return [ item for sublist in list_ for item in sublist ]

    def trim_tree(self, tree, prob_dict, topk):
        """
        Trims the tree to the top k candidates.

        Args:
            tree (List[str]): List of candidate strings.
            prob_dict (dict): Dictionary of probabilities of the candidates.
            topk (int): Number of top candidates to be retained.
        
        Returns:
            List[str]: List of top k candidates.
        """
        output = []
        probs = [ prob_dict[x] for x in tree ]
        candidates = [
            Node(item, item.replace(" ", ""), probs[idx]) for idx, item in enumerate(tree)
        ]
        for key, group in itertools.groupby(candidates, key=lambda x: x.characters):
            sorted_group = sorted(list(group), key=lambda x: x.score)
            trimmed_group = sorted_group[0:topk]
            trimmed_group = [x.hypothesis for x in trimmed_group]
            output.extend(trimmed_group)
        return output

    def run(self, dataset, topk=20, steps=13):
        """
        Runs the beamsearch algorithm on the provided dataset.

        Args:
            dataset (List[str]): List of initial candidate strings.
            topk (int): Number of top candidates to be retained in each step. Default is 20.
            steps (int): Number of steps to run the algorithm. Default is 13.
        
        Returns:
            ProbabilityDictionary: Dictionary of final probabilities of the candidates.
        """
        tree = dataset
        prob_dict = {}
        for i in range(steps):
            tree = self.next_step(tree)
            tree = self.reshape_tree(tree, self.gpu_batch_size)
            prob_dict = self.update_probabilities(tree, prob_dict)
            tree = self.flatten_list(tree)
            tree = self.trim_tree(tree, prob_dict, topk)
        return ProbabilityDictionary(prob_dict)


from hashformers.beamsearch.minicons_lm import MiniconsLM

class GPT2LM(MiniconsLM):
    """A Language Model (LM) scorer using GPT2.

    This class utilizes the PaddedGPT2LMScorer for scoring sentences.

    Args:
        model_name_or_path (str): Name or path of the model to be used.
        device (str): The device to run the model on. Default is 'cuda'.
        gpu_batch_size (int): The batch size for GPU processing. Default is 20.
    """
    def __init__(self, model_name_or_path, device='cuda', gpu_batch_size=20):
        super().__init__(
            model_name_or_path=model_name_or_path,
            device=device,
            gpu_batch_size=gpu_batch_size,
            model_type='IncrementalLMScorer'
        )


#!/usr/bin/python

#
# Modeler.py
# 
# Copyright (c) 2016, Bogazici University. See the file
# COPYRIGHT.txt in the software or at http://tabilab.cmpe.boun.edu.tr/projects/hashtag_segmentation/copyright.html
# 
# This file is part of BOUN Hashtag Segmentor (see http://tabilab.cmpe.boun.edu.tr/projects/hashtag_segmentation),
# and is free software, licensed under the GNU Library General Public License, Version 2, June 1991
# (in the distribution as file licence.html, and also available at
# http://tabilab.cmpe.boun.edu.tr/projects/hashtag_segmentation/licence.html).
# 
# Arda Celebi, TABI Labs, CMPE, Bogazici University, 
#

class Modeler(object):
    
    hashtagSegmentor = None
    t = 0
    totals = 0
    totalh = 0
    p = 0
    r = 0
    n = 0
    
    modelerParams = {}
    
    def __init__(self):
        pass
    
    def loadParameters(self, args):
        leftoverArgs = []
        for arg in args:
            if self.loadParameter(arg) == False:
                leftoverArgs.append(arg)    
        return leftoverArgs
    
    def loadParameter(self, param):
        pass
    
    def getRunCode(self):
        return ""
    
    def train(self, featureFile):
        pass
    
    def segmentHashtag(self, hashtag):
        pass
    
    def segmentFile(self, fileToSegment, featureFileName, params):
        pass
    
    def calculateScore(self, testFile, params):
        pass
    
    def loadModelerParams(self, params):
        pass
    
    def test(self, testFile, featureFileName, params):
        (acc, precision, recall, fscore) = self.calculateScore(testFile, featureFileName, params)
    
        print("MAXENT ACC %f PRE %f REC %f F1 %f\n" % (acc, precision, recall, fscore))
        
    def isFeatureOn(self, feature):
        return False
    
    def reset(self):
        self.t = 0
        self.totals = 0
        self.totalh = 0
        self.p = 0
        self.r = 0
        self.n = 0
    
    def countEntry(self, segmented, trueSegmentation):
        sw = segmented.split(' ')
        hw = trueSegmentation.split(' ')
        for s in sw:
            for h in hw:
                if s == h:
                    self.p = self.p + 1
                    break
        for h in hw:
            for s in sw:
                if s == h:
                    self.r = self.r + 1
                    break
                
        self.totals = self.totals + len(sw)
        self.totalh = self.totalh + len(hw)
        self.n += 1
        
        if segmented == trueSegmentation:
            self.t += 1
        
    def calculatePrecision(self):
        if self.totals > 0:
            return ((float)(self.p*100)/(float)(self.totals))
        return 0
    
    def calculateRecall(self):
        if self.totalh > 0:
            return ((float)(self.r*100)/(float)(self.totalh))
        return 0
    
    def calculateFScore(self):
        precision = self.calculatePrecision()
        recall = self.calculateRecall()
        
        if precision+recall > 0:
            return 2*precision*recall/(precision+recall)
        return 0
    
    def calculateAccuracy(self):
        if self.n > 0:
            return ((float)(100*self.t)/(float)(self.n))
        return 0



from hashformers.segmenter.base_segmenter import BaseSegmenter
import re

class RegexWordSegmenter(BaseSegmenter):
    """
    A subclass of BaseSegmenter which uses regex rules to perform segmentation.

    Args:
        regex_rules (list of str, optional): List of regex rules used for segmentation. 
        If none are provided, it uses the default rule: [r'([A-Z]+)'].

    Attributes:
        regex_rules (list of _sre.SRE_Pattern): List of compiled regex rules used for segmentation.
    """
    def __init__(self,regex_rules=None):
        """
        Initializes the RegexWordSegmenter with given or default regex rules and compiles them.
        """
        if not regex_rules:
            regex_rules = [r'([A-Z]+)']
        self.regex_rules = [
            re.compile(x) for x in regex_rules
        ]

    def segment_word(self, rule, word):
        """
        Segments a word based on a given regex rule.

        Args:
            rule (_sre.SRE_Pattern): The compiled regex rule to be used for segmentation.
            word (str): The word to be segmented.

        Returns:
            str: The segmented word.
        """
        return rule.sub(r' \1', word).strip()

    def segmentation_generator(self, word_list):
        """
        A generator that iterates over the word list and yields segmented words 
        based on the regex rules.

        Args:
            word_list (list of str): The list of words to be segmented.

        Yields:
            str: The segmented word.
        """
        for rule in self.regex_rules:
            for idx, word in enumerate(word_list):
                yield self.segment_word(rule, word)

    def segment(self, inputs: list[str], **kwargs):
        """
        Segments a list of strings based on the regex rules. Before segmentation, 
        the inputs are preprocessed using the inherited preprocess method.

        Args:
            inputs (List[str]): The list of strings to be segmented.
            **kwargs: Arbitrary keyword arguments for the inherited preprocess method.

        Returns:
            list of str: The segmented inputs.
        """
        inputs = super().preprocess(inputs, **kwargs)
        return list(self.segmentation_generator(inputs))


from hashformers.evaluation.modeler import Modeler
import pandas as pd
import copy
import numpy as np

def evaluate_df(
    df, 
    gold_field="gold", 
    segmentation_field="segmentation"
):
  """
  Evaluates the given dataframe based on the gold_field and segmentation_field and returns various metric values.

  This function creates a new column "truth_value" in the dataframe by comparing gold_field and segmentation_field.
  It then sorts the dataframe by gold_field and "truth_value" and retains the first row for each gold_field group.
  The metrics calculated include F1 score, accuracy, recall and precision.

  Args:
      df (pandas.DataFrame): The dataframe to be evaluated.
      gold_field (str, optional): The field in the dataframe used as the 'truth' field for evaluation. Defaults to "gold".
      segmentation_field (str, optional): The field in the dataframe used as the 'prediction' field for evaluation. Defaults to "segmentation".
  
  Returns:
      dict: A dictionary containing F1 score, accuracy, recall, and precision metrics.
  """
  evaluator = Modeler()

  df["truth_value"] = df[gold_field].combine(
    df[segmentation_field],
    lambda x,y: x == y
  )

  df = df\
    .sort_values(
      by=[
        gold_field, 
        "truth_value"
      ], 
      ascending=False)\
    .groupby(gold_field)\
    .head(1)

  records = df.to_dict("records")
  for row in records:
    evaluator.countEntry(
      row[segmentation_field],
      row[gold_field]
    )
  metrics = {
      "f1": evaluator.calculateFScore(),
      "acc": evaluator.calculateAccuracy(),
      "recall": evaluator.calculateRecall(),
      "precision": evaluator.calculatePrecision()
  }
  return metrics

def filter_top_k(
    input_df, 
    k, 
    gold_field="hashtag", 
    score_field="score",
    segmentation_field="segmentation",
    fill=False):
  """
  Filters the top k rows of the input_df for each group defined by the gold_field. 

  The function sorts the input_df by score_field in ascending order and retains the first k rows for each group.
  If fill option is set to True, it also clones the records based on the length of each group. 

  Args:
      input_df (pandas.DataFrame): The input dataframe to filter.
      k (int): The number of top records to retain for each group.
      gold_field (str, optional): The field used to define groups in the dataframe. Defaults to "hashtag".
      score_field (str, optional): The field used to sort the dataframe. Defaults to "score".
      segmentation_field (str, optional): The field used if the fill option is set to True. Defaults to "segmentation".
      fill (bool, optional): Whether to clone the records based on the length of each group. Defaults to False.

  Returns:
      pandas.DataFrame: The filtered dataframe.
  """
  df = copy.deepcopy(input_df)
  
  df = df\
    .sort_values(by=score_field, ascending=True)\
    .groupby(gold_field)\
    .head(k)

  if fill:
    df["group_length"] = df.groupby(gold_field)[segmentation_field].transform(len)
    df["group_length"] = df["group_length"] * -1 + k + 1
    len_array = df["group_length"].values
    
    df = df.drop(columns=["group_length"])
    records = np.array(df.to_dict("records"))
    cloned_records = list(np.repeat(records, len_array))
    df = pd.DataFrame(cloned_records)
    
    df = df\
      .sort_values(by=score_field, ascending=True)\
      .groupby(gold_field)\
      .head(k)

    length = df.groupby(gold_field).size().values
    assert (length == k).all()
  
  return df


def read_experiment_dataset(data, dataset, model):
    """
    Reads and returns the dataset for a given model from a collection of datasets.

    The function filters the data based on the dataset and model parameters, converts the filtered data into a pandas 
    DataFrame and returns it.

    Args:
        data (list of dicts): The collection of datasets. Each element is a dictionary which must contain "dataset" and 
        "model" keys along with a "data" key which contains the actual data.
        dataset (str): The name of the dataset to read.
        model (str): The name of the model for which the dataset needs to be read.

    Returns:
        pandas.DataFrame: The selected dataset as a dataframe.
    """
    selected_data = [ 
      x for x in data if x["dataset"]==dataset \
        and x["model"]==model
      ][0]["data"]
    output = pd.DataFrame(selected_data)
    return output


from setuptools import find_packages, setup

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name='hashformers',
    version='2.2.0',
    author='Ruan Chaves Rodrigues',
    author_email='ruanchave93@gmail.com',
    description='Word segmentation with transformers',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ruanchaves/hashformers",
    packages=find_packages('src'),
    package_dir={'': 'src'},
    python_requires=">=3.8",
    install_requires=[
        "minicons",
        "twitter-text-python",
        "pandas"
    ],
    extras_require={
        "spacy": ["spacy>=3.0.0"]
    },
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Text Processing :: Linguistic",
    ],
    keywords="word-segmentation hashtag nlp transformers spacy",
)
